"""
Infrastructure pages.
"""

#c Copyright 2008-2018, the GAVO project
#c
#c This program is free software, covered by the GNU GPL.  See the
#c COPYING file in the source distribution.


import datetime
import io
import os
import time

import pkg_resources
from nevow import inevow
from nevow import rend
from nevow import static
from nevow import url
from twisted.web import http

from gavo import base
from gavo import registry
from gavo import svcs
from gavo import utils
from gavo.base import meta
from gavo.votable import V
from gavo.web import caching
from gavo.web import grend

# prefer globally installed version to what we deliver
try:
	import rjsmin
except ImportError:
	from gavo.imp import rjsmin

class LoginPage(rend.Page, grend.GavoRenderMixin):
	"""a page that logs people in or out.

	You should usually give a nextURL parameter in the context, the page
	the user is returned to afte login.

	If the user is already authenticated, this will do a logout (by
	sending a 403).
	"""
	def __init__(self, ctx):
		rend.Page.__init__(self)
		self.request = inevow.IRequest(ctx)
		self.nextURL = self.request.args.get("nextURL", ["/"])[0]

	def render_nextURL(self, ctx, data):
		return ctx.tag(href=self.nextURL)

	def render_iflogged(self, ctx, data):
		if self.request.getUser():
			return ctx.tag
		return ""
	
	def render_ifnotlogged(self, ctx, data):
		if not self.request.getUser():
			return ctx.tag
		return ""

	def data_loggedUser(self, ctx, data):
		return self.request.getUser()

	def renderHTTP(self, ctx):
		relogging = base.parseBooleanLiteral(utils.getfirst(
			self.request.args, "relog", default="False"))
		if self.request.getUser():  # user is logged in...
			if relogging: # ...and wants to log out: show login dialog...
				raise svcs.Authenticate()
			else:   # ...and has just logged in: forward to destination
				return url.URL.fromContext(ctx).click(self.nextURL)
		else:  # user is not logged in
			if relogging:  #...but was and has just logged out: forward to dest
				return url.URL.fromContext(ctx).click(self.nextURL)
			else: # ... and wants to log in.
				raise svcs.Authenticate()

	docFactory = svcs.loadSystemTemplate("loginout.html")


def _replaceConfigStrings(srcPath, registry):
	src = open(srcPath).read().decode("utf-8")
	src = src.replace("__site_path__", base.getConfig("web", "nevowRoot"))
	src = src.replace("__site_url__", os.path.join(
		base.getConfig("web", "serverURL")+base.getConfig("web", "nevowRoot")))
	return src.encode("utf-8")


class TemplatedPage(grend.CustomTemplateMixin, grend.ServiceBasedPage):
	"""a "server-wide" template.

	For now, they all are based on the dc root service.
	"""
	checkedRenderer = False
	def __init__(self, ctx, fName):
		self.customTemplate = fName
		grend.ServiceBasedPage.__init__(self, ctx,
			base.caches.getRD(registry.SERVICELIST_ID
			).getById("root"))
		self.metaCarrier = meta.MetaMixin()
		self.metaCarrier.setMetaParent(self.service)
		self.metaCarrier.setMeta("dateUpdated", 
			utils.formatISODT(
				datetime.datetime.fromtimestamp(os.path.getmtime(fName))))


def minifyJS(ctx, path):
	"""returns javascript in path minified.

	You can turn off auto-minification by setting [web] jsSource to True;
	that's sometimes convenient while debugging the javascript.

	If jsSource is false (the default), changes to javascript are only picked
	up on a server reload.
	"""
	with open(path) as f:
		if base.getConfig("web", "jsSource"):
			return f.read()
		else:
			return rjsmin.jsmin(f.read())


def expandTemplate(ctx, fName):
	"""renders fName as a template on the root service.
	"""
	return TemplatedPage(ctx, fName)


class StaticFile(rend.Page):
	"""a file from the file system, served pretty directly.

	Some of these static files can be javascript (that's usually minified
	on the fly) or nevow templates.  These we want to cache.  For everything
	else, we don't win anything by caching, but, when we serve large data,
	we can DoS ourselves.  So, we decide manually whether to cache.

	The caches are bound to an RD passed in as cacheRD at construction time.
	For system resources, that should be getRD(registry.SERVICELIST_ID).

	On-the-fly processing is based on certain magic mime types:
	text/nevow-template is rendered and application/javascript is minified.
	"""
	defaultType = "application/octet-stream"

	magicMimes = {
		"text/nevow-template": expandTemplate,
		"application/javascript": minifyJS,
	}

	def __init__(self, fName, cacheRD):
		self.fName, self.cacheRD = fName, cacheRD

	def getMimeType(self):
		ext = os.path.splitext(self.fName)[-1]
		return static.File.contentTypes.get(ext, self.defaultType)

	def renderPlain(self, request):
		static.FileTransfer(
			open(self.fName), os.path.getsize(self.fName), request)

	def renderHTTP(self, ctx):
		request = inevow.IRequest(ctx)
		modStamp = max(self.cacheRD.loadedAt, os.path.getmtime(self.fName))
		if request.setLastModified(modStamp) is http.CACHED:
			return ''

		if not os.path.isfile(self.fName):
			raise svcs.ForbiddenURI("Only plain files are served here")

		mime = self.getMimeType()
		if mime in self.magicMimes:
			cache = base.caches.getPageCache(self.cacheRD.sourceId)
			cachedRes = cache.get(self.fName)
			if cachedRes is not None and cachedRes.creationStamp>modStamp:
				return cachedRes
			caching.instrumentRequestForCaching(request,
				caching.enterIntoCacheAs(self.fName, cache))
			return (self.magicMimes[mime](ctx, self.fName))

		else:
			request.setHeader("content-type", mime)
			self.renderPlain(request)

		return (request.deferred)


########### Begin Debian Fallbacks: If no local jquery and friends are
# present (as happens when the official package is used)
# several javascript libraries are taken from where Debian packages
# put them.  The URI path parts DaCHS sees them under are mapped to
# the absolute path names or resource instances in DEBIAN_FALLBACKS.

class GavoJquery(rend.Page):
	"""synthesises a query-gavo.js from Debian packages.
	"""
	INSTALL_ROOT = "/usr/share/javascript/"
	PARTS = ["jquery/jquery.js", 
		io.BytesIO("""
				(function( factory ) {
				if ( typeof define === "function" && define.amd ) {

					// AMD. Register as an anonymous module.
					define([ "jquery" ], factory );
				} else {

					// Browser globals
					factory( jQuery );
				}
			}(function( $ ) {

			$.ui = $.ui || {};

			var version = $.ui.version = "1.12.1";
			"""),
		"jquery-ui/ui/widget.js",
		"jquery-ui/ui/position.js",
		"jquery-ui/ui/data.js",
		"jquery-ui/ui/disable-selection.js",
		"jquery-ui/ui/focusable.js",
		"jquery-ui/ui/form-reset-mixin.js",
		"jquery-ui/ui/jquery-1-7.js",
		"jquery-ui/ui/keycode.js",
		"jquery-ui/ui/labels.js",
		"jquery-ui/ui/scroll-parent.js",
		"jquery-ui/ui/tabbable.js",
		"jquery-ui/ui/unique-id.js",
		"jquery-ui/ui/widgets/mouse.js",
		io.BytesIO("""
			var plugin = $.ui.plugin = {
				add: function( module, option, set ) {
					var i,
						proto = $.ui[ module ].prototype;
					for ( i in set ) {
						proto.plugins[ i ] = proto.plugins[ i ] || [];
						proto.plugins[ i ].push( [ option, set[ i ] ] );
					}
				},
				call: function( instance, name, args, allowDisconnected ) {
					var i,
						set = instance.plugins[ name ];

					if ( !set ) {
						return;
					}

					if ( !allowDisconnected && ( !instance.element[ 0 ].parentNode ||
							instance.element[ 0 ].parentNode.nodeType === 11 ) ) {
						return;
					}

					for ( i = 0; i < set.length; i++ ) {
						if ( instance.options[ set[ i ][ 0 ] ] ) {
							set[ i ][ 1 ].apply( instance.element, args );
						}
					}
				}
			};
			"""),
		"jquery-ui/ui/widgets/draggable.js",
		"jquery-ui/ui/widgets/resizable.js",

		io.BytesIO("}));"),
		]

	def __init__(self):
		self.content = None
		rend.Page.__init__(self)

	def _getContent(self):
		parts = []
		for part in self.PARTS:
			if hasattr(part, "read"):
				parts.append(part.read())
			else:
				with open(os.path.join(self.INSTALL_ROOT, part)) as f:
					parts.append(f.read())
		stuff = "\n".join(parts)

		self.modStamp = time.time()
		if base.getConfig("web", "jsSource"):
			return stuff
		else:
			return rjsmin.jsmin(stuff)
		
	def renderHTTP(self, ctx):
		if self.content is None:
			self.content = self._getContent()

		request = inevow.IRequest(ctx)
		if request.setLastModified(self.modStamp) is http.CACHED:
			return ''

		return self.content


DEBIAN_FALLBACKS = {
	'js/jquery-gavo.js': GavoJquery(),
	'js/jquery.flot.js': 
		"/usr/share/javascript/jquery-flot/jquery.flot.js",
}

########### End Debian Fallbacks

class StaticServer(rend.Page):
	"""is a server for various static files.

	This is basically like static.File, except

		- we don't do directory listings
		- we don't bother with ranges
		- we look for each file in a user area and then in the system area.
	"""
	def __init__(self):
		rend.Page.__init__(self)
		self.userPath = utils.ensureOneSlash(
			os.path.abspath(os.path.join(base.getConfig("webDir"), "nv_static")))
		self.systemPath = utils.ensureOneSlash(
			os.path.abspath(pkg_resources.resource_filename('gavo', "resources/web")))

	def renderHTTP(self, ctx):
		raise svcs.UnknownURI("What did you expect here?")

	def locateChild(self, ctx, segments):
		relPath = "/".join(segments)
		associatedRD = base.caches.getRD(registry.SERVICELIST_ID)

		path = os.path.abspath(self.userPath+relPath)
		if os.path.exists(path):
			if not path.startswith(self.userPath):
				raise svcs.ForbiddenURI(
					"%s is not located in the static user directory"%path)
			return StaticFile(path, associatedRD), ()

		path = os.path.abspath(self.systemPath+relPath)
		if os.path.exists(path):
			if not path.startswith(self.systemPath):
				raise svcs.ForbiddenURI(
					"%s is not located in the static system directory"%path)
			return StaticFile(path, associatedRD), ()

		if relPath in DEBIAN_FALLBACKS:
			res = DEBIAN_FALLBACKS[relPath]
			if isinstance(res, basestring):
				return StaticFile(res, associatedRD), ()
			else:
				return res, ()

		raise svcs.UnknownURI("No matching file,"
			" neither built-in nor user-provided")

	processors = {
		".shtml": _replaceConfigStrings,
	}


class RobotsTxt(rend.Page):
	"""A page combining some built-in robots.txt material with etc/robots.txt
	if it exists.
	"""
	builtin = utils.fixIndentation("""
		Disallow: /login
		Disallow: /seffe
		""", "")

	def _getContent(self):
		content = self.builtin
		try:
			with open(os.path.join(base.getConfig("webDir"), "robots.txt")) as f:
				content = content+"\n"+f.read()
		except IOError:
			pass
		return content

	def renderHTTP(self, ctx):
		request = inevow.IRequest(ctx)
		request.setHeader("content-type", "text/plain")
		return self._getContent()

	def locateChild(self, segments):
		return None


class ServiceUnavailable(rend.Page):
	"""A page to be rendered in emergencies.

	Essentially, this is a 503 with a text taken from stateDir/MAINT.

	Root checks for the presence of that file before returning this
	page, so (ignoring race conditions) this page assumes it's there.
	"""
	def renderHTTP(self, ctx):
		request = inevow.IRequest(ctx)
		request.setResponseCode(503)
		request.setHeader("retry-after", "3600")

		with open(os.path.join(base.getConfig("stateDir"), "MAINT")) as f:
			try:
				maintText = f.read().decode("utf-8")
			except IOError:
				maintText = "Unspecified problem"
		
		vot = V.VOTABLE[
			V.RESOURCE(type="results")[
				V.INFO(name="QUERY_STATUS", value="ERROR")[
					maintText]
			]]

		request.setHeader("content-type", "text/xml")
		return ("<?xml-stylesheet href='/static"
			"/xsl/dalierror-to-html.xsl' type='text/xsl'?>"
			+vot.render())
